!
! Test the function SG03CD employing a Lyapunov Equation 
!
PROGRAM TESTLYAP
    IMPLICIT NONE 

    ! Parameters
    INTEGER             N, EXAMPLE, IONE
    PARAMETER           ( N = 500, EXAMPLE = 3, IONE = 1 ) 
    DOUBLE PRECISION    MONE, ONE, ZERO
    PARAMETER           ( MONE = -1.0d0, ONE = 1.0d0, ZERO = 0.0d0 )
    

    ! Matrices
    DOUBLE PRECISION    A(N,N), E(N,N), X(N,N), Y(N,N), U(N,N), B(N,N)
    DOUBLE PRECISION    Q(N,N), Z(N,N), Y2(N,N), A2(N,N), E2(N,N) 
    DOUBLE PRECISION    DWORK(2*N*N+4488)
    DOUBLE PRECISION    ALPHAR(N), ALPHAI(N), BETA(N)
    INTEGER             IWORK(N**2+66)
    LOGICAL             BWORK(N) 
    

    ! Scalars and additional Variables 
    INTEGER             NR(2)
    DOUBLE PRECISION    DPAR(2)
    INTEGER             IPAR(1)
    LOGICAL             VEC(8)
    INTEGER             ON, OM
    CHARACTER           NOTE(70)
    INTEGER             LDWORK, INFO, I
    DOUBLE PRECISION    SCALE, FERR,SEP, NRMX, FERR2,SEP2, NRMY,NRMR


    ! External Functions and Subroutines 
    EXTERNAL            BB03AD, SG03CD, DLACPY, DAXPY, DLANGE, SG03AD, DGEMM
    DOUBLE PRECISION    DLANGE

    INTRINSIC DABS

    ! .. Program start ..

    WRITE(*,*) '*************************************************'
    WRITE (*,*) "SG03CD Example"

    ! Generate Example 
    NR(1) = 4
    NR(2) = EXAMPLE
    DPAR(1) = 40
    IPAR(1) = N 
    LDWORK = 2*N**2+4488

    CALL BB03AD('N', NR, DPAR, IPAR, VEC, ON , OM, E, N, A, N, Y, N, B,N,X, N, U, N,NOTE, DWORK, LDWORK, INFO)

    WRITE(*,*) 'Generated Example: ', NOTE
    WRITE(*,*) 'of dimension ', ON 
    IF ( ON .NE. N ) THEN 
        WRITE(*,*) 'Desired Dimension does not match. Abort'
        STOP
    END IF 

    NRMX = DLANGE('F', N, N , X, N, DWORK)
    NRMY = DLANGE('F', N, N , Y, N, DWORK)

    CALL DLACPY("All",N,N, Y, N, Y2, N)
    CALL DLACPY("All",N,N, A, N, A2, N)
    CALL DLACPY("All",N,N, E, N, E2, N)

    WRITE(*,*) '*************************************************'
    WRITE(*,*) '        A**T * X * E + E**T * X * A = Y'

    !
    ! First test 
    !
    WRITE(*,*) '*************************************************'
    WRITE(*,*) 'Solve the Equation including QZ factorization'
    CALL SG03CD('Continuous','X', 'N', 'N','U',N, A, N, E, N, Q, N, Z, N, Y, N,&
        & SCALE,SEP, FERR, ALPHAR, ALPHAI, BETA, IWORK, DWORK, LDWORK, BWORK, INFO) 

    WRITE(*,*) 'INFO = ', INFO

    ! Check Residual 
    CALL DLACPY('All', N , N, Y2, N, U, N)
    CALL DGEMM('T', 'N', N, N, N, ONE, A2, N , Y, N, ZERO, B, N)
    CALL DGEMM('N', 'N', N, N, N, MONE, B, N, E2, N, ONE, U, N)
    CALL DGEMM('T', 'N', N, N, N, ONE, E2, N , Y, N, ZERO, B, N)
    CALL DGEMM('N', 'N', N, N, N, MONE, B, N, A2, N, ONE, U, N)
    NRMR = DLANGE('F', N, N, U, N, DWORK)
    WRITE(*,*) 'Rel. Residual = ', NRMR/NRMY 

    ! Check Forward Error
    DO I = 1, N
        CALL DAXPY(N, MONE, X(1,I), IONE, Y(1,I), IONE)
    END DO 
    WRITE(*,*) 'Forward Error:', DLANGE('F',N,N,Y,N,DWORK)/NRMX


    !
    ! Solve with already factorized A, E 
    !
    WRITE(*,*) '*************************************************'
    CALL DLACPY('All', N , N , Y2, N, Y, N)
    WRITE(*,*) 'Solve the Equation with existing QZ factorization'
    CALL SG03CD('Continuous','X', 'F', 'N','U',N, A, N, E, N, Q, N, Z, N, Y, N,&
        & SCALE,SEP, FERR, ALPHAR, ALPHAI, BETA, IWORK, DWORK, LDWORK, BWORK, INFO) 

    WRITE(*,*) 'INFO = ', INFO

    ! Check Residual 
    CALL DLACPY('All', N , N, Y2, N, U, N)
    CALL DGEMM('T', 'N', N, N, N, ONE, A2, N , Y, N, ZERO, B, N)
    CALL DGEMM('N', 'N', N, N, N, MONE, B, N, E2, N, ONE, U, N)
    CALL DGEMM('T', 'N', N, N, N, ONE, E2, N , Y, N, ZERO, B, N)
    CALL DGEMM('N', 'N', N, N, N, MONE, B, N, A2, N, ONE, U, N)
    NRMR = DLANGE('F', N, N, U, N, DWORK)
    WRITE(*,*) 'Rel. Residual = ', NRMR/NRMY 

    ! Check Forward Error
    DO I = 1, N
        CALL DAXPY(N, MONE, X(1,I), IONE, Y(1,I), IONE)
    END DO 
    WRITE(*,*) 'Forward Error:', DLANGE('F',N,N,Y,N,DWORK)/NRMX

   
    !
    ! Compute the separation only 
    !
    WRITE(*,*) '*************************************************'
    WRITE(*,*) 'Compute the separation only.'
    CALL SG03CD('Continuous','S', 'F', 'N','U',N, A, N, E, N, Q, N, Z, N, Y, N,&
        & SCALE,SEP, FERR, ALPHAR, ALPHAI, BETA, IWORK, DWORK, LDWORK, BWORK, INFO) 


    WRITE(*,*) 'INFO = ', INFO
    WRITE(*,*) 'Separation = ', SEP

    WRITE(*,*) 'Compute SG03AD reference'
    CALL SG03AD('Continuous','S', 'F', 'N','U',N, A, N, E, N, Q, N, Z, N, Y, N,&
        & SCALE,SEP2, FERR2, ALPHAR, ALPHAI, BETA, IWORK, DWORK, LDWORK, INFO) 


    WRITE(*,*) 'INFO = ', INFO
    WRITE(*,*) 'Separation = ', SEP2
    WRITE(*,*) 'Separation diff = ', DABS(SEP-SEP2)


    !
    ! Compute Both 
    ! 
    WRITE(*,*) '*************************************************'
    WRITE(*,*) 'Compute the separation and solution.'
    CALL DLACPY('All', N , N , Y2, N, Y, N)
    CALL SG03CD('Continuous','B', 'F', 'N','U',N, A, N, E, N, Q, N, Z, N, Y, N,&
        & SCALE,SEP, FERR, ALPHAR, ALPHAI, BETA, IWORK, DWORK, LDWORK, BWORK, INFO) 


    WRITE(*,*) 'INFO = ', INFO
    WRITE(*,*) 'Separation    = ', SEP
    DO I = 1, N
        CALL DAXPY(N, MONE, X(1,I), IONE, Y(1,I), IONE)
    END DO 
    WRITE(*,*) 'Forward Error = ', DLANGE('F',N,N,Y,N,DWORK)/NRMX
    WRITE(*,*) 'FERR estimate = ', FERR 

    
    CALL DLACPY('All', N , N , Y2, N, Y, N)

    WRITE(*,*) 'Compute SG03AD reference'
    CALL SG03AD('Continuous','B', 'F', 'N','U',N, A, N, E, N, Q, N, Z, N, Y, N,&
        & SCALE,SEP2, FERR2, ALPHAR, ALPHAI, BETA, IWORK, DWORK, LDWORK, INFO) 


    WRITE(*,*) 'INFO = ', INFO
    WRITE(*,*) 'Separation = ', SEP2
    DO I = 1, N
        CALL DAXPY(N, MONE, X(1,I), IONE, Y(1,I), IONE)
    END DO 
    WRITE(*,*) 'Forward Error = ', DLANGE('F',N,N,Y,N,DWORK)/NRMX
    WRITE(*,*) 'FERR estimate = ', FERR2

    WRITE(*,*) 'Separation diff = ', DABS(SEP-SEP2)
    WRITE(*,*) 'FERR est.  diff = ', DABS(FERR-FERR2)

    WRITE(*,*) '*************************************************'

    

    !
    ! Test the transposed case 
    !
    
    WRITE(*,*) ''
    WRITE(*,*) '*************************************************'
    WRITE(*,*) '        A * X * E**T + E * X * A**T = Y'
    CALL DLACPY("All", N, N, A2, N, A, N)
    CALL DLACPY("All", N, N, E2, N, E, N)


    ! Build the right hand side 
    
    CALL DGEMM('N','N', N , N, N, ONE, A, N, X, N, ZERO, B, N)
    CALL DGEMM('N','T', N , N, N, ONE, B, N, E, N, ZERO, Y, N)
    CALL DGEMM('N','N', N , N, N, ONE, E, N, X, N, ZERO, B, N)
    CALL DGEMM('N','T', N , N, N, ONE, B, N, A, N, ONE, Y, N)
    CALL DLACPY('All', N, N, Y, N, Y2, N) 
    NRMY = DLANGE('F', N, N , Y, N, DWORK)

    !
    ! First test 
    !
    WRITE(*,*) '*************************************************'
    WRITE(*,*) 'Solve the Equation including QZ factorization'
    CALL SG03CD('Continuous','X', 'N', 'T','U',N, A, N, E, N, Q, N, Z, N, Y, N,&
        & SCALE,SEP, FERR, ALPHAR, ALPHAI, BETA, IWORK, DWORK, LDWORK, BWORK, INFO) 

    WRITE(*,*) 'INFO = ', INFO

    ! Check Residual 
    CALL DLACPY('All', N , N, Y2, N, U, N)
    CALL DGEMM('N', 'N', N, N, N, ONE, A2, N , Y, N, ZERO, B, N)
    CALL DGEMM('N', 'T', N, N, N, MONE, B, N, E2, N, ONE, U, N)
    CALL DGEMM('N', 'N', N, N, N, ONE, E2, N , Y, N, ZERO, B, N)
    CALL DGEMM('N', 'T', N, N, N, MONE, B, N, A2, N, ONE, U, N)
    NRMR = DLANGE('F', N, N, U, N, DWORK)
    WRITE(*,*) 'Rel. Residual = ', NRMR/NRMY 

    ! Check Forward Error
    DO I = 1, N
        CALL DAXPY(N, MONE, X(1,I), IONE, Y(1,I), IONE)
    END DO 
    WRITE(*,*) 'Forward Error:', DLANGE('F',N,N,Y,N,DWORK)/NRMX

    !
    ! Solve with already factorized A, E 
    !
    WRITE(*,*) '*************************************************'
    CALL DLACPY('All', N , N , Y2, N, Y, N)
    WRITE(*,*) 'Solve the Equation with existing QZ factorization'
    CALL SG03CD('Continuous','X', 'F', 'T','U',N, A, N, E, N, Q, N, Z, N, Y, N,&
        & SCALE,SEP, FERR, ALPHAR, ALPHAI, BETA, IWORK, DWORK, LDWORK, BWORK, INFO) 

    WRITE(*,*) 'INFO = ', INFO

    ! Check Residual 
    CALL DLACPY('All', N , N, Y2, N, U, N)
    CALL DGEMM('N', 'N', N, N, N, ONE, A2, N , Y, N, ZERO, B, N)
    CALL DGEMM('N', 'T', N, N, N, MONE, B, N, E2, N, ONE, U, N)
    CALL DGEMM('N', 'N', N, N, N, ONE, E2, N , Y, N, ZERO, B, N)
    CALL DGEMM('N', 'T', N, N, N, MONE, B, N, A2, N, ONE, U, N)
    NRMR = DLANGE('F', N, N, U, N, DWORK)
    WRITE(*,*) 'Rel. Residual = ', NRMR/NRMY 

    ! Check Forward Error
    DO I = 1, N
        CALL DAXPY(N, MONE, X(1,I), IONE, Y(1,I), IONE)
    END DO 
    WRITE(*,*) 'Forward Error:', DLANGE('F',N,N,Y,N,DWORK)/NRMX

   
    !
    ! Compute the separation only 
    !
    WRITE(*,*) '*************************************************'
    WRITE(*,*) 'Compute the separation only.'
    CALL SG03CD('Continuous','S', 'F', 'T','U',N, A, N, E, N, Q, N, Z, N, Y, N,&
        & SCALE,SEP, FERR, ALPHAR, ALPHAI, BETA, IWORK, DWORK, LDWORK, BWORK, INFO) 


    WRITE(*,*) 'INFO = ', INFO
    WRITE(*,*) 'Separation = ', SEP

    WRITE(*,*) 'Compute SG03AD reference'
    CALL SG03AD('Continuous','S', 'F', 'T','U',N, A, N, E, N, Q, N, Z, N, Y, N,&
        & SCALE,SEP2, FERR2, ALPHAR, ALPHAI, BETA, IWORK, DWORK, LDWORK, INFO) 


    WRITE(*,*) 'INFO = ', INFO
    WRITE(*,*) 'Separation = ', SEP2
    WRITE(*,*) 'Separation diff = ', DABS(SEP-SEP2)


    !
    ! Compute Both 
    ! 
    WRITE(*,*) '*************************************************'
    WRITE(*,*) 'Compute the separation and solution.'
    CALL DLACPY('All', N , N , Y2, N, Y, N)
    CALL SG03CD('Continuous','B', 'F', 'T','U',N, A, N, E, N, Q, N, Z, N, Y, N,&
        & SCALE,SEP, FERR, ALPHAR, ALPHAI, BETA, IWORK, DWORK, LDWORK, BWORK, INFO) 


    WRITE(*,*) 'INFO = ', INFO
    WRITE(*,*) 'Separation    = ', SEP
    DO I = 1, N
        CALL DAXPY(N, MONE, X(1,I), IONE, Y(1,I), IONE)
    END DO 
    WRITE(*,*) 'Forward Error = ', DLANGE('F',N,N,Y,N,DWORK)/NRMX
    WRITE(*,*) 'FERR estimate = ', FERR 

    
    CALL DLACPY('All', N , N , Y2, N, Y, N)

    WRITE(*,*) 'Compute SG03AD reference'
    CALL SG03AD('Continuous','B', 'F', 'T','U',N, A, N, E, N, Q, N, Z, N, Y, N,&
        & SCALE,SEP2, FERR2, ALPHAR, ALPHAI, BETA, IWORK, DWORK, LDWORK, INFO) 


    WRITE(*,*) 'INFO = ', INFO
    WRITE(*,*) 'Separation = ', SEP2
    DO I = 1, N
        CALL DAXPY(N, MONE, X(1,I), IONE, Y(1,I), IONE)
    END DO 
    WRITE(*,*) 'Forward Error = ', DLANGE('F',N,N,Y,N,DWORK)/NRMX
    WRITE(*,*) 'FERR estimate = ', FERR2

    WRITE(*,*) 'Separation diff = ', DABS(SEP-SEP2)
    WRITE(*,*) 'FERR est.  diff = ', DABS(FERR-FERR2)

    WRITE(*,*) '*************************************************'

END PROGRAM 
